package CFG;
use strict;
use bytes;

#
# Generate an S19 file containing specified configuration values.
# line, generate S19 file lines with checksums, etc., to load those
# values at that address.
#
# NOTE: Includes dummy header and footer records.  Remove these if
# making a composite S19 file.
# 

my ($base, $fmt, $lineno, $line, $addrlen, @used, $origin, %polllast, %_req);
my ($suppress);


#
# Use top 2 bits of widget type to encode a background colour.
#
my %pale = (W => 0, R => (1 << 6), Y => (2 << 6), B => (3 << 6));

my %len = (
	U16 => 2, S16 => 2, F16 => 2, U8 => 1, S8 => 1
);

#
# Configuration pages live at these addresses.  Page 0 is the default.
#
my %base = (
	0 => 0x1000,
	1 => 0xf400, 2 => 0xf4c0, 3 => 0xf580, 4 => 0xf640, 5 => 0xf700
);

#
# Return index of first free widget slot in panel.
#
sub firstfree
{
	for (0 .. 15)
	{
		return $_ if $used[$_] == 0;
	}
	return -1;
}

sub hexsum
{
	my ($hstr) = @_;

	my $sum = 0;
	for (my $i = 0; $i < length($hstr); $i+= 2)
	{
		$sum += hex(substr($hstr, $i, 2));
	}
	$sum;
}

sub rgb
{
	my ($r, $g, $b) = @_;
	my $v = (($r & 0xf8)<<8) + (($g & 0xfc) << 3) + (($b & 0xf8) >> 3);
	return(chr($v & 0xff) . chr($v >> 8));
}

sub zpad
{
	my ($s, $l) = @_;

	my ($p);
	if (($p = $l - length($s)))
	{
		$s .= "\0" x $p;
	}
	$s;
}

sub mks19
{
	my ($base, $lineno, $line) = @_;
	my ($sum, $len, $str);

	$len = length($line);
	die "$lineno:invalid hex line" unless $line =~ /^[0-9A-Fa-f]+$/;
	die "$lineno:odd line length" unless ($len & 1) == 0;
	$len /= 2;
	$str = sprintf "$fmt", $addrlen + $len + 1, $base;
	$str .= $line;
	$sum = ~hexsum(substr($str, 2)) & 0xff;
	printf("%s%02X\n", $str, $sum);
	$len;
}

#
# Convert a string to hexadecimal and pad with trailing values to a
# fixed length.
#
sub hexpad
{
	my ($str, $len, $pad) = @_;
	my $rv = '';

	for (my $i =0; $i < length($str); $i++)
	{
		$rv .= sprintf "%02X", ord(substr($str,$i,1));
	}
	$len *= 2;
	while (length($rv) < $len)
	{
		$rv .= $pad;
	}
	$rv;
}

sub checkpollcycle
{
	my ($v) = @_;
	for my $i ((2,4,8,16,32,64,128))
	{
		return 1 if $v == $i;
	}
	die "Illegal pollcycle $v\n";
}

sub cleanup
{
	# Clean up module globals so module can be called repeatedly.
	undef $lineno;
	undef $line;
	undef @used;
	undef $origin;
	undef %polllast;
	undef %_req;
}

sub nohdr
{
	$suppress = 1;
}

my @crctbl = (
   0x00000000,0x77073096,0xEE0E612C,0x990951BA,0x076DC419,0x706AF48F,0xE963A535,
   0x9E6495A3,0x0EDB8832,0x79DCB8A4,0xE0D5E91E,0x97D2D988,0x09B64C2B,0x7EB17CBD,
   0xE7B82D07,0x90BF1D91,0x1DB71064,0x6AB020F2,0xF3B97148,0x84BE41DE,0x1ADAD47D,
   0x6DDDE4EB,0xF4D4B551,0x83D385C7,0x136C9856,0x646BA8C0,0xFD62F97A,0x8A65C9EC,
   0x14015C4F,0x63066CD9,0xFA0F3D63,0x8D080DF5,0x3B6E20C8,0x4C69105E,0xD56041E4,
   0xA2677172,0x3C03E4D1,0x4B04D447,0xD20D85FD,0xA50AB56B,0x35B5A8FA,0x42B2986C,
   0xDBBBC9D6,0xACBCF940,0x32D86CE3,0x45DF5C75,0xDCD60DCF,0xABD13D59,0x26D930AC,
   0x51DE003A,0xC8D75180,0xBFD06116,0x21B4F4B5,0x56B3C423,0xCFBA9599,0xB8BDA50F,
   0x2802B89E,0x5F058808,0xC60CD9B2,0xB10BE924,0x2F6F7C87,0x58684C11,0xC1611DAB,
   0xB6662D3D,0x76DC4190,0x01DB7106,0x98D220BC,0xEFD5102A,0x71B18589,0x06B6B51F,
   0x9FBFE4A5,0xE8B8D433,0x7807C9A2,0x0F00F934,0x9609A88E,0xE10E9818,0x7F6A0DBB,
   0x086D3D2D,0x91646C97,0xE6635C01,0x6B6B51F4,0x1C6C6162,0x856530D8,0xF262004E,
   0x6C0695ED,0x1B01A57B,0x8208F4C1,0xF50FC457,0x65B0D9C6,0x12B7E950,0x8BBEB8EA,
   0xFCB9887C,0x62DD1DDF,0x15DA2D49,0x8CD37CF3,0xFBD44C65,0x4DB26158,0x3AB551CE,
   0xA3BC0074,0xD4BB30E2,0x4ADFA541,0x3DD895D7,0xA4D1C46D,0xD3D6F4FB,0x4369E96A,
   0x346ED9FC,0xAD678846,0xDA60B8D0,0x44042D73,0x33031DE5,0xAA0A4C5F,0xDD0D7CC9,
   0x5005713C,0x270241AA,0xBE0B1010,0xC90C2086,0x5768B525,0x206F85B3,0xB966D409,
   0xCE61E49F,0x5EDEF90E,0x29D9C998,0xB0D09822,0xC7D7A8B4,0x59B33D17,0x2EB40D81,
   0xB7BD5C3B,0xC0BA6CAD,0xEDB88320,0x9ABFB3B6,0x03B6E20C,0x74B1D29A,0xEAD54739,
   0x9DD277AF,0x04DB2615,0x73DC1683,0xE3630B12,0x94643B84,0x0D6D6A3E,0x7A6A5AA8,
   0xE40ECF0B,0x9309FF9D,0x0A00AE27,0x7D079EB1,0xF00F9344,0x8708A3D2,0x1E01F268,
   0x6906C2FE,0xF762575D,0x806567CB,0x196C3671,0x6E6B06E7,0xFED41B76,0x89D32BE0,
   0x10DA7A5A,0x67DD4ACC,0xF9B9DF6F,0x8EBEEFF9,0x17B7BE43,0x60B08ED5,0xD6D6A3E8,
   0xA1D1937E,0x38D8C2C4,0x4FDFF252,0xD1BB67F1,0xA6BC5767,0x3FB506DD,0x48B2364B,
   0xD80D2BDA,0xAF0A1B4C,0x36034AF6,0x41047A60,0xDF60EFC3,0xA867DF55,0x316E8EEF,
   0x4669BE79,0xCB61B38C,0xBC66831A,0x256FD2A0,0x5268E236,0xCC0C7795,0xBB0B4703,
   0x220216B9,0x5505262F,0xC5BA3BBE,0xB2BD0B28,0x2BB45A92,0x5CB36A04,0xC2D7FFA7,
   0xB5D0CF31,0x2CD99E8B,0x5BDEAE1D,0x9B64C2B0,0xEC63F226,0x756AA39C,0x026D930A,
   0x9C0906A9,0xEB0E363F,0x72076785,0x05005713,0x95BF4A82,0xE2B87A14,0x7BB12BAE,
   0x0CB61B38,0x92D28E9B,0xE5D5BE0D,0x7CDCEFB7,0x0BDBDF21,0x86D3D2D4,0xF1D4E242,
   0x68DDB3F8,0x1FDA836E,0x81BE16CD,0xF6B9265B,0x6FB077E1,0x18B74777,0x88085AE6,
   0xFF0F6A70,0x66063BCA,0x11010B5C,0x8F659EFF,0xF862AE69,0x616BFFD3,0x166CCF45,
   0xA00AE278,0xD70DD2EE,0x4E048354,0x3903B3C2,0xA7672661,0xD06016F7,0x4969474D,
   0x3E6E77DB,0xAED16A4A,0xD9D65ADC,0x40DF0B66,0x37D83BF0,0xA9BCAE53,0xDEBB9EC5,
   0x47B2CF7F,0x30B5FFE9,0xBDBDF21C,0xCABAC28A,0x53B39330,0x24B4A3A6,0xBAD03605,
   0xCDD70693,0x54DE5729,0x23D967BF,0xB3667A2E,0xC4614AB8,0x5D681B02,0x2A6F2B94,
   0xB40BBE37,0xC30C8EA1,0x5A05DF1B,0x2D02EF8D
);

sub crc32
{
	my ($str) = @_;
	my ($v);

	$v = 0xffffffff;
	for (0..(length($str)-1))
	{
		$v = ($v >> 8) ^ $crctbl[($v ^ ord(substr($str,$_,1))) & 0xff];
	}
	$v ^= 0xffffffff;
	sprintf "%c%c%c%c", $v>>24, ($v >> 16)&0xff, ($v>>8)&0xff, $v&0xff;
}


#
# Old style (pre CRC32) OutputChannels request
#
sub ms2_oldrequest
{
	my ($offset, $nb) = @_;
	my (%req);

	#
	# Hack -- keep these values for later checks
	#
	$_req{hdr} = 0;
	$_req{first} = $offset;
	$_req{buflen} = $nb;
	$req{query_vals} = sprintf("r\x00\x07\x00%c\x00%c", $offset, $nb);
	$req{resp_len} = $nb;
	%req;
}

sub ms2_request
{
	my ($offset, $nb) = @_;
	my (%req);

	#
	# Hack -- keep these values for later checks
	#
	$_req{hdr} = 3;
	$_req{buflen} = $nb;
	# command length 0007; command r0<off>0<len> 
	$req{query_vals} = "r\x00\x07\x00" . chr($offset) . "\x00" . chr($nb);
	# append CRC32 checksum
	$req{query_vals} .= crc32 $req{query_vals};
	$req{query_vals} = "\x00\x07" . $req{query_vals};
	# response length is requested bytes + 2 of length, 1 of
	# status, 4 of crc.
	$req{resp_len} = $nb + 7;
	%req;
}

my %conf;


sub
binshort
{
	my ($v) = @_;

	return(chr($v & 0xff) . chr($v >> 8));
}


#
# Information in type field:
#   Bits 6:7 -- pale colour index
#   Bit  5   -- inverse (black bg, coloured fg)
#   Bit  4   -- 1 => refresh on every poll
#   Bit  3   -- available
#   Bit  0:2 -- widget type
#
sub
packtype
{
	my ($w, $t) = @_;
	my ($always, $inverse);
	if (!defined($pale{$w->{bg}}))
	{
		warn "Unknown bg colour '$w->{bg}' -- using 'W'\n";
		$w->{bg} = 'w';
	}
	if (exists $w->{inverse})
	{
		$inverse = $w->{inverse} ? 1 << 5 : 0;
	}
	else
	{
		$inverse = 0;
	}
	$always = defined($polllast{$w->{field}}) ? 1 << 4 : 0;
	return chr($t + $always + $inverse+ $pale{$w->{bg}});
}

#
# Blank widget uses bits[6,7]=0,1,2  for left,centre,right alignment;
#   bits[4,5]=0,1,2 for top, middle, bottom alignment.
# NOTE: bit of a hack.  Reusing ALWAYS bit (0x10) so blank widgets
# with a "middle" vertical justification will be visited on every
# poll.
#
sub
packblanktype
{
	my ($w) = @_;
	my %halign = (left => 0, centre => 0x40, center => 0x40,
			right => 0x80);
	my %valign = (top => 0x0, middle => 0x10, bottom => 0x20);
	my $t = 0;
	if (exists($w->{halign}))
	{
		die "Unknown halignment '$w->{halign}'\n" unless
				exists($halign{$w->{halign}});
		$t += $halign{$w->{halign}};
	}
	if (exists($w->{valign}))
	{
		die "Unknown valignment '$w->{valign}'\n" unless
				exists($valign{$w->{valign}});
		$t += $valign{$w->{valign}};
	}
	$t;
}

sub
parseblank
{
	my ($w, $hr, $idx, $buf) = @_;
	my ($tmp, $tmp2);

	# Either 5 bytes of zeroes, or a zero followed by up to 4
	# bytes of text label.
	if (exists($w->{label}))
	{
		$tmp = chr(packblanktype($w));
		if (!exists($w->{inverse}) or $w->{inverse} == 0)
		{
			$tmp2 = $w->{label};
		}
		else
		{
			$tmp2 = chr(ord($w->{label}) | 0x80) .
					substr($w->{label},1);
		}
		$tmp .= sprintf("%.4s", $tmp2);
		$tmp = hexpad($tmp, 5, '00');
		substr(${$buf}, 10*$idx, 10) = $tmp;
	}
	$used[$idx] = 1;
	$origin |= 1 << $idx;
}

sub
parsetext
{
	my ($w, $hr, $idx, $buf) = @_;
	my ($tmp);

	$tmp = packtype($w, 1);
	die "Unknown handle $w->{field}\n" unless exists $hr->{$w->{field}};
	$tmp .= chr($hr->{$w->{field}}->[0]);	# poll field index
	# Bump handle use count
	$hr->{$w->{field}}->[1]++;
	# 3 byte label
	$tmp .= sprintf("%3.3s", $w->{label});
	$tmp = hexpad($tmp, 5, '00');
	#
	# Store in buffer
	#
	$used[$idx] = 1;
	$origin |= 1 << $idx;
	substr(${$buf}, 10*$idx, 10) = $tmp;
}


sub
parseledbar
{
	my ($w, $hr, $idx, $buf) = @_;
	my ($tmp);

	die "Wrong side ledbar\n" if $idx & 1;
	die "Overlap ledbar in slot $idx\n" if $used[$idx+1];
	$tmp = packtype($w, 2);
	die "Unknown handle $w->{field}\n" unless exists $hr->{$w->{field}};
	die "Bad range for $w->{field}\n" if $w->{minval} >= $w->{maxval};
	$tmp .= chr($hr->{$w->{field}}->[0]);	# poll field index
	# Bump handle use count
	$hr->{$w->{field}}->[1]++;
	$tmp .= binshort($w->{minval});		# lowest value
	#
	# Complexities of next line
	#  1. config specifies maxval, dash wants range
	#  2. 8/7 because we want rightmost LED lit as soon as value reached
	#  3. &0xfff8 because of bug in some firmware using RRC not  RRA
	$tmp .= binshort(int(8*($w->{maxval} - $w->{minval})/7) & 0xfff8);
	$tmp .= chr(5);				# TODO LED colour transitions
	$tmp .= "\0\0\0";			# pad out second slot
	$tmp = hexpad($tmp, 10, '00');
	$origin |= 1 << $idx;
	$used[$idx] = 1;
	$used[$idx + 1] = 1;
	substr(${$buf}, 10*$idx, 20) = $tmp;
}

sub
parsebar
{
	my ($w, $hr, $idx, $buf) = @_;
	my ($tmp);

	die "Wrong side bar\n" if $idx & 1;
	die "Overlap bar in slot $idx\n" if $used[$idx+1] +
			$used[$idx+2] + $used[$idx+3];
	$tmp = packtype($w, 3);
	die "Unknown handle $w->{field}\n" unless exists $hr->{$w->{field}};
	$tmp .= chr($hr->{$w->{field}}->[0]);	# poll field index
	# Bump handle use count
	$hr->{$w->{field}}->[1]++;
	$tmp .= binshort($w->{minval});		# lowest value
	$tmp .= binshort($w->{maxval});		# highest value
	$tmp .= binshort($w->{blue});		# blue part of gauge
	$tmp .= binshort($w->{red});		# red part of gauge
	$tmp = zpad($tmp, 17);
	$tmp .= sprintf("%3.3s", $w->{label});	# label
	$tmp = hexpad($tmp, 20, '00');		# padding
	$origin |= 1 << $idx;
	$used[$idx] = 1;
	$used[$idx + 1] = 1;
	$used[$idx + 2] = 1;
	$used[$idx + 3] = 1;
	substr(${$buf}, 10*$idx, 40) = $tmp;
}

sub
parserangebar
{
	my ($w, $hr, $idx, $buf) = @_;
	my ($tmp);

	die "Wrong side rangebar\n" if $idx & 1;
	die "Overlap rangebar in slot $idx\n" if $used[$idx+1] +
			$used[$idx+2] + $used[$idx+3];
	$tmp = packtype($w, 4);
	die "Unknown handle $w->{field}\n" unless exists $hr->{$w->{field}};
	$tmp .= chr($hr->{$w->{field}}->[0]);	# poll field index
	# Bump handle use count
	$hr->{$w->{field}}->[1]++;
	$tmp .= binshort($w->{minval});		# lowest value
	$tmp .= binshort($w->{maxval});		# highest value
	die "Unknown handle $w->{lofield}\n" unless exists $hr->{$w->{lofield}};
	$tmp .= chr($hr->{$w->{lofield}});
	$hr->{$w->{lofield}}->[1]++;
	die "Unknown handle $w->{hifield}\n" unless exists $hr->{$w->{hifield}};
	$tmp .= chr($hr->{$w->{hifield}});
	$hr->{$w->{hifield}}->[1]++;
	$tmp .= binshort($w->{tick});		# blue part of gauge
	$tmp = zpad($tmp, 17);
	$tmp .= sprintf("%3.3s", $w->{label});	# label
	$tmp = hexpad($tmp, 20, '00');	# convert and pad
	$origin |= 1 << $idx;
	$used[$idx] = 1;
	$used[$idx + 1] = 1;
	$used[$idx + 2] = 1;
	$used[$idx + 3] = 1;
	substr(${$buf}, 10*$idx, 40) = $tmp;
}

sub
parsetriled
{
	my ($w, $hr, $idx, $buf) = @_;
	my ($tmp);

	die "Wrong side triled\n" if $idx & 1;
	die "Overlap triled in slot $idx\n" if $used[$idx+1];
	$tmp = packtype($w, 5);
	die "Unknown handle $w->{field}\n" unless exists $hr->{$w->{field}};
	$tmp .= chr($hr->{$w->{field}}->[0]);	# poll field index
	# Bump handle use count
	$hr->{$w->{field}}->[1]++;
	$tmp .= sprintf("%3.3s", $w->{label});	# label
	$tmp .= chr($w->{cmap});		# colour map
	$tmp .= binshort($w->{loval});		# low threshold
	$tmp .= binshort($w->{hival});		# hi threshold
	$tmp = hexpad($tmp, 10, '00');	# convert and pad
	$origin |= 1 << $idx;
	$used[$idx] = 1;
	$used[$idx + 1] = 1;
	substr(${$buf}, 10*$idx, 20) = $tmp;
}

sub
parseseg7
{
	my ($w, $hr, $idx, $buf) = @_;
	my ($tmp);

	die "Too low seg7\n" if $idx > 13;
	die "Overlap seg7 in slot $idx\n" if $used[$idx+2];
	$tmp = packtype($w, 6);
	die "Unknown handle $w->{field}\n" unless exists $hr->{$w->{field}};
	$tmp .= chr($hr->{$w->{field}}->[0]);	# field index
	# Bump handle use count
	$hr->{$w->{field}}->[1]++;
	if (exists($w->{cmap}))
	{
		die "seg7 cmap must have loval and hival in slot $idx\n"
				unless exists($w->{loval}) &&
				exists($w->{hival});
		$tmp .= chr($w->{cmap});
		$tmp .= chr(0) . chr(0);
		if (($idx & 1))
		{
			$tmp .= chr(0);		# alignment
		}
		$tmp .= binshort($w->{loval});
		$tmp .= binshort($w->{hival});
	}
	$tmp = hexpad($tmp, 10, '00');
	$origin |= 1 << $idx;
	$used[$idx] = 1;
	$used[$idx + 2] = 1;
	# Top half, then bottom half (not that there's much in there)
	substr(${$buf}, 10*$idx, 10) = substr($tmp,0,10);
	substr(${$buf}, 10*($idx+2), 10) = substr($tmp,10,10);
}

#
# Work out an appropriate scaler for a set of grid values.
#
sub gridscaler
{
	my ($ar) = @_;
	my ($pow2, $maxrng);

	#
	# Make sure buckets are in ascending order
	#
	$maxrng = 0;
	for (1..$#{$ar})
	{
		die "GUIDEGRID buckets must be in ascending order" if
				$ar->[$_-1] >= $ar->[$_];
		#if ($_ > 3 and $ar->[$_] - $ar->[$_ - 3] > $maxrng)
		#{
		#	$maxrng = $ar->[$_] - $ar->[$_ - 3];
		#}
		#
		# Really just interested in the range from [0:max]
		#
		$maxrng = $ar->[$_];
	}
	#
	# How many shifts until it's fractional, then scale up by 2^7
	#
	$pow2 = int(log($maxrng)/log(2)) - 7;
	$pow2 = 0 if $pow2 < 0;
	$pow2;
}

sub packgridvals
{
	my ($w, $hr, $ftype) = @_;
	my ($tmp, $nbuckets, $shifts);

	$nbuckets = $#{$w->{"${ftype}buckets"}};
	die "GUIDEGRID needs ${ftype}buckets array\n"
			unless exists $w->{"${ftype}buckets"};
	die "${ftype}buckets array must have 16 or fewer members\n"
			if $nbuckets > 15;
	$shifts = gridscaler($w->{"${ftype}buckets"});
	$tmp .= chr($shifts);
	#
	# Arbitrary check
	#
	warn "GUIDEGRID ${ftype}range probably too big\n"
			if ($w->{"${ftype}range"} >> $shifts) > 180;
	warn "GUIDEGRID ${ftype}range probably too small\n"
			if ($w->{"${ftype}range"} >> $shifts) < 10;
	$tmp .= chr($w->{"${ftype}range"} >> $shifts);
	foreach (@{$w->{"${ftype}buckets"}})
	{
		die "packgridvals:$_ shifted $shifts is out of range\n"
				if $_>> $shifts <= 0 or $_ >> $shifts > 255;
		$tmp .= chr($_ >> $shifts);
	}
	while ($nbuckets < 15)
	{
		$tmp .= chr(0xff);
		$nbuckets++;
	}
	$tmp;
}


sub parseguidegrid
{
	my ($w, $hr, $idx, $buf) = @_;
	my ($tmp, $shifts);

	die "Too low guidegrid\n" if $idx > 3;
	die "Wrong side guidegrid\n" if $idx & 1;
	for (0..11)
	{
		die "Overlap guidegrid in slot $idx\n" if $used[$idx+$_];
		$used[$idx+$_] = 1;
	}
	#
	# TODO allow for black guide ticks
	#
	$tmp = packtype($w, 7);
	#
	# Store struct grad for horizontal grid spec
	#
	die "Unknown handle $w->{field}\n" unless exists $hr->{$w->{field}};
	$tmp .= chr($hr->{$w->{field}}->[0]);	# field index
	$tmp .= packgridvals($w, $hr, "h");
	#
	# Store struct grad for vertical grid spec
	#
	$tmp .= chr($hr->{$w->{vfield}}->[0]);
	$tmp .= packgridvals($w, $hr, "v");
	die "Unknown handle $w->{vfield}\n" unless exists $hr->{$w->{vfield}};
	$tmp = hexpad($tmp, 60, '00');
	$origin |= 1 << $idx;
	substr(${$buf}, 10*$idx, 120) = $tmp;
}


my %parser = (
	BLANK => \&parseblank,
	TEXT => \&parsetext,
	LEDBAR => \&parseledbar,
	BAR => \&parsebar,
	RANGEBAR => \&parserangebar,
	TRILED => \&parsetriled,
	SEG7 => \&parseseg7,
	GUIDEGRID => \&parseguidegrid,
);

sub compiles19
{
	my ($buf, $pbuf, %phandle, $idx);
	# Each array pointer says [#bytes in target, encoding type]
	# Encodings are: 's' -- string with leading byte count
	#                'b' -- string without count
	#                'i' -- single byte integer
	my (%desc) = (
		palette => [8, 'b'],	#W-RGB, R-RGB, Y-RGB, B-RGB
		expect_id => [24,'s'],
		query_vals => [16,'s'],
		resp_len => [1,'i'],
		pollcycle => [1, 'i'],
		bg => [4, 'b'],		# RGB, RGB
		blduty => [2, 'i'],	# day, night [0:100]
		txdelay => [1, 'i'],
		pad => [1, 'i'],	# bug patch (txdelay read as word)
		orientation => [1, 'i'],# MADCTL value (packs divider too)
		divider => [1, 'i'],	# SPI clock divider
		npoll => [1, 'i'],	# n variables being accumulated
		);
	#
	# Default value for these.
	#
	$conf{palette} = rgb(0xff,0xff,0xff) . rgb(0xff,0x90,0x90) .
			rgb(0xff,0xff,0x40) . rgb(0xa0,0xff,0xff)
			unless exists $conf{palette};
	$conf{orientation} = 0 unless exists $conf{orientation};
	$conf{divider} = 1 unless exists $conf{divider};
	$conf{npoll} = $#{$conf{poll}} + 1;
	die "Divider must be [1:4]\n" if $conf{divider} < 1 or
			$conf{divider} > 4;
	$conf{pad} = 0;
	#
	# Divider is packed in unused bits in orientation/MADCTL
	#
	$conf{orientation} |= $conf{divider} - 1;
	#
	# Deal with the simple string configuration values
	#
	foreach (qw/palette expect_id query_vals resp_len
			pollcycle bg blduty txdelay pad orientation npoll/)
	{
		die "No '$_'\n" unless exists $conf{$_};
		my $str = '';
		my $pad = '00';
		if ($desc{$_}->[1] eq 's')
		{
			$str = chr(length($conf{$_})) . $conf{$_};
		}
		elsif ($desc{$_}->[1] eq 'b')
		{
			$str = sprintf "%s", $conf{$_};
			$pad = '20';	# pad with spaces
		}
		elsif ($desc{$_}->[1] eq 'i')
		{
			if ($desc{$_}->[0] > 1)
			{
				foreach my $c (@{$conf{$_}})
				{
					$str .= chr($c);
				}
			}
			else
			{
				$str = chr($conf{$_});
			}
		}
		else
		{
			die "unknown format description $desc{$_}->[1]\n";
		}
		$buf .= hexpad($str, $desc{$_}->[0], $pad);
	}
	#
	# Make sure the pollcycle value is a acceptable.
	#
	checkpollcycle $conf{pollcycle};
	#
	# Turn the standard types and formats into numbers.
	#
	my (%method, %type);
	my $i = 0;
	foreach my $t (qw/mean min max last/)	# 0,1,2,3
	{
		$method{$t} = $i++;
	}
	$i = 0;
	foreach my $t (qw/U8 U16 S8 S16/)	# 0,4,8,c
	{
		$type{$t} = 4*$i++;
	}
	$type{F16} = 0xc + 0x80;		# S16 + bit7 (Temperature bit)

	#
	# Buffer the polled values.  Count is stored at npoll above.
	# This is the polled values themselves.  Description at top of
	# format.s
	#
	die "Not allowed more than 16 poll entries\n" if $#{$conf{poll}} > 15;
	$idx = 0;
	foreach my $f (@{$conf{poll}})
	{
		my $str = '';
		my $toss = exists($f->{o_discard}) ? $f->{o_discard} : 0;
		my ($p) = $f->{o_prec};
		die "no variable handle on var $idx\n" unless $f->{handle};
		die "duplicate poll handle $f->{handle}\n"
				if exists $phandle{$f->{handle}};
		$phandle{$f->{handle}} = [$idx++, 0];
		die "illegal precision $p\n" unless $p > -4 and $p < 4;
		die "illegal discards $toss\n" if $toss > 3 or $toss < 0;
		die "unknown type $f->{o_type}\n"
				unless exists $type{$f->{o_type}};
		die "unknown method $f->{o_method}\n"
				unless exists $method{$f->{o_method}};
		die "no offset specified for $f->{handle}\n"
				unless exists $f->{o_buf};
		#
		# Remember which variables will update their widgets every
		# sample.  This will be packed into the widget type byte.
		#
		if ($f->{o_method} eq "last")
		{
			$polllast{$f->{handle}} = 1;
		}
		if (($f->{o_buf} & 1) and $f->{o_type} =~ /16/)
		{
			die "poll $f->{handle}: word on odd alignment\n";
		}
		#
		# Make sure offset is within buffer.
		#
		if ($f->{o_buf} < 0)
		{
			die "o_buf value for $f->{handle} < 0\n";
		}
		my ($end) = $f->{o_buf} + $len{$f->{o_type}};
		if (exists($_req{buflen}))
		{
			# header length known
			if ($end > $_req{buflen} + $_req{hdr})
			{
				die "$f->{handle} goes past end of read data\n";
			}
		}
		else
		{
			if ($end > $conf{resp_len})
			{
				die "o_buf value $f->{o_buf} outside response limits\n";
			}
		}
		#
		# Precision is coded as absolute value in 2 lsb and
		# sign in msb.
		#
		$p = ($p < 0) ? 0x80 | -$p : $p;
		$str = chr($f->{o_buf}) . chr(16*$toss + $type{$f->{o_type}} +
				$method{$f->{o_method}}) . chr($p);
		$pbuf .= hexpad($str, 3, '00');
	}
	#
	# Now compile widgets to be shown
	#
	$idx = 0;
	my $tbuf = '00' x 80;
	foreach my $s (@{$conf{show}})
	{
		die "Too many 'show' entries\n" if firstfree() == -1;
		$s->{bg} = 'W' unless defined($s->{bg});
		die "$s->{type} unknown\n" unless exists($parser{$s->{type}});
		&{$parser{$s->{type}}}($s, \%phandle, firstfree(), \$tbuf);
	}
	die "must fill all 16 slots\n" unless firstfree() == -1;
	#
	# Tell about any handles which weren't used.
	#
	foreach (keys %phandle)
	{
		next if $phandle{$_}->[1];
		warn "**WARNING** Unused poll handle '$_'\n", $_;
	}
	#
	# Some spare room in first page of info flash.  Last 2-bytes
	# is the mask of widget origins.  Then append widget and
	# variable definitions.
	#
	$buf .= sprintf('0000%02X%02X', $origin & 0xff, $origin >> 8) .
			$tbuf . $pbuf;
	#
	# Pad to the end of 192 bytes with zeroes
	#
	while (length($buf) < 384)
	{
		$buf .= '00';
	}
	$buf;
}


$base = 0x1000;			# store config in info flash
$fmt = "S1%02X%04X";		# always 16-bit addresses
$addrlen = 2;

#
# Set the configuration page base address for the S19 output.
#
sub setpage
{
	my ($pg) = @_;

	die "Page must be [0:5]\n" unless exists $base{$pg};
	($base) = $base{$pg};
}

sub mkconf
{
	my ($hr) = @_;

	print "S00700006461736858\n" unless $suppress;
	%conf = %$hr;
	@used = (0) x 16;
	$origin = 0;
	my $buf = compiles19();
	die "Wrong conf length\n" unless length($buf) == 384;
	for (my $i=0; $i < 384; $i += 32)
	{
		my $str = substr($buf, $i, 32);
		$lineno++;
		$base += mks19($base, $lineno, $str);
	}
	print "S9030000FC\n" unless $suppress;
}

1;
