/*
 * Software to run on a laptop to reflash the dash.
 */

#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <ctype.h>

typedef unsigned char U8;

struct ctl
{
	const char *fn;		/* S19 file name */
	int lineno;		/* line # in S19 file */
	const char *port;	/* serial device name */
	FILE *cf;		/* config file stdio pointer */
	int serfd;		/* serial port file descriptor */
	unsigned int mask;	/* bit mask of tainted segments */
	unsigned char this;	/* block most recently asked for */
	unsigned char next;	/* next block to be asked for */
	U8 blk[256];		/* blocks that have been loaded */
	U8 mem[65536];		/* binary configuration read from cf */
};

/*
 * Return the bit mask corresponding to the segment containing a given address.
 * Return -1 if the address is outside [0x1000-0x10c0]+[0xc000-0xf800].
 */
static unsigned int
segment_mask(struct ctl *ctl, int addr)
{
	int blk;
	/*
	 * Addresses 0x1000-0x10bf are writable info mem.  64 byte segments.
	 */
	if (addr < 0x10c0 && addr >= 0x1000)
	{
		/*
		 * info flash is blocks 252, 253, 254
		 */
		blk = 252 + ((addr & 0xc0) >> 6);
		ctl->blk[blk] = 1;
		/*
		 * Top three bits for info flash.
		 *  1000 -> 1<<28, 1040 -> 1<<29, 1080 -> 1<<30.
		 */
		return(1 << (28 + ((addr & 0xc0) >> 6)));
	}
	/*
	 * Addresses 0xc000-0xffff are writable flash.	512 byte segments.
	 */
	if (addr < 0xf800 && addr >= 0xc000)
	{
		/*
		 * If any block is loaded, mark the whole segment
		 */
		blk = ((addr - 0xc000) >> 6) & ~0x07;
		memset(ctl->blk + blk, 1, 8);
		/*
		 * Lowest 28 bits for program flash.
		 *  c000 -> 1, c200 -> 2, c400 -> 3, etc.
		 */
		return(1 << ((addr - 0xc000) >> 9));
	}
	return((unsigned int)(-1));
}



static volatile int terminate;

/*
 * Send contents of buf to the given serial port.  Echo characters to the
 * screen as they are sent.
 */
int
to_serial(int len, char *buf, int serfd)
{
	write(serfd, buf, len);
	return(0);
}

/*
 * Send a 64-byte block of data to the peer, complete with two-byte checksum.
 * Block numbering scheme is a little eccentric.  Numbers n in [0:223]
 * transmit memory at 0xc000+64*n, if n is in [253:255], the info flash
 * segment at 0x1000+64*(255-n) is sent, otherwise it's an error.
 */
static int
xmit_block(struct ctl *ctl, int blockno)
{
	int off;
	int i;
	int sum;
	U8 c;
	int baseaddr;
	U8 *base;

	if (blockno >= 252 && blockno < 255)
	{
		printf("Request for config segment #%d\n", blockno - 252);
		baseaddr = 0x1000 + 64 * (blockno - 252);
	}
	else if (blockno < 224)
	{
		printf("Request for flash block #%d\n", blockno);
		baseaddr = 0xc000 + 64 * blockno;
	}
	else
	{
		fprintf(stderr, "Request for illegal block %d\n", blockno);
		return(-1);
	}
	if (blockno == ctl->this)
	{
		printf("*** repeat ***\n");
	}
	else if (blockno == ctl->next)
	{
		/*
		 * Update next expected request.
		 */
		ctl->this = blockno;
		ctl->next = 255;
		for (i = blockno + 1; i < 255; i++)
		{
			if (ctl->blk[i])
			{
				ctl->next = i;
				break;
			}
		}
	}
	else
	{
		printf("Request for %d out of sequence; expected %d\n",
				blockno, ctl->next);
		return(-1);
	}
				
	base = &ctl->mem[baseaddr];
	sum = 0;
	for (i = 0; i < 64; i++)
	{
		c = *base++;
		/*
		 * Send the byte to the serial port.
		 */
		(void)to_serial(1, &c, ctl->serfd);
		/*
		 * Rotate left one bit
		 */
		sum <<= 1;
		if (sum & 0x00010000)
		{
			sum &= 0xffff;
			sum++;
		}
		/*
		 * XOR in the new byte.
		 */
		sum ^= c;
	}
	/*
	 * Send the checksum in bigendian order.
	 */
	c = sum >> 8;
	to_serial(1, &c, ctl->serfd);
	c = sum & 0xff;
	to_serial(1, &c, ctl->serfd);
	return(0);
}


/*
 * Serial comms event loop
 */
static int
dispatch(struct ctl *ctl)
{
	fd_set infd;
	U8 buf;
	int n;
	enum {
		NEU,
		GOT_B,
	} state = NEU;

	while (!terminate)
	{
		n = read(ctl->serfd, &buf, 1);
		if (n < 0 && errno != EINTR)
		{
			perror("read");
		}
		switch (state)
		{
		case NEU:
			printf("'%c'", buf);
			fflush(stdout);
			if (buf == 'Q')
			{
				to_serial(20, "MASTER MASTER MASTER",
						ctl->serfd);
			}
			else if (buf == 'M')
			{
				/* 32 bits of data flash mask */
				to_serial(4, (char *)&ctl->mask, ctl->serfd);
			}
			else if (buf == 'B')
			{
				/* expect block id in next byte */
				state = GOT_B;
			}
			else if (buf == 'X')
			{
				terminate = 1;
			}
			else
			{
				fprintf(stderr, "Unexpected char %02x\n", buf);
			}
			break;

		case GOT_B:
			/*
			 * Send appropriate block.
			 */
			xmit_block(ctl, buf);
			state = NEU;
			break;
		}
	}
	return(0);
}


static int
open_serial(const char *dev)
{
	int fd;
	int mode;
	struct termios tio;

	/*
	 * Open in nonblocking mode to begin with, so we can set CLOCAL.
	 */
	if ((fd = open(dev, O_RDWR | O_NOCTTY | O_NONBLOCK)) < 0)
	{
		return(-1);
	}
	if (flock(fd, LOCK_EX | LOCK_NB) != 0)
	{
		fprintf(stderr, "%s:device file locked\n", dev);
		return(-1);
	}
	if (tcgetattr(fd, &tio) < 0)
	{
		close(fd);
		return(-1);
	}
	cfsetspeed(&tio, B115200);
	tio.c_iflag |= IGNBRK;
	tio.c_iflag &= ~(ISTRIP | ICRNL | IXON | IXOFF);
	tio.c_lflag &= ~(ICANON | ISIG | IEXTEN | ECHO);
	tio.c_oflag &= ~(OPOST|ONLCR|OCRNL|ONOCR|ONLRET);
	tio.c_cflag |= (CLOCAL | CREAD | CS8);
	tio.c_cflag &= ~(CRTSCTS | CSTOPB | PARENB);
	tio.c_cc[VMIN] = 1;		/* > 1 chars */
	tio.c_cc[VTIME] = 1;		/* blocking */
	if (tcsetattr(fd, 0, &tio) < 0)
	{
		close(fd);
		return(-1);
	}
	/*
	 * CLOCAL has been set.  Switch to blocking I/O.
	 */
	mode = fcntl(fd, F_GETFL, NULL);
	if (mode < 0)
	{
		close(fd);
		perror("F_GETFL");
		return(-1);
	}
	mode &= ~(O_NONBLOCK | O_NDELAY);
	if (fcntl(fd, F_SETFL, mode) < 0)
	{
		close(fd);
		perror("F_SETFL");
		return(-1);
	}
	/*
	 * Get rid of anything accumulated in the buffers.
	 */
	tcflush(fd, TCIOFLUSH);
	return(fd);
}


static int
hexdigit(int c)
{
	if (c >= '0' && c <= '9')
	{
		return(c - '0');
	}
	if (c >= 'A' && c <= 'F')
	{
		return(c - 'A' + 10);
	}
	if (c >= 'a' && c <= 'f')
	{
		return(c - 'a' + 10);
	}
	return(-1);
}


static int
hexbyte(const char *ptr)
{
	int ms;
	int ls;

	if ((ms = hexdigit(ptr[0])) < 0)
	{
		return(-1);
	}
	if ((ls = hexdigit(ptr[1])) < 0)
	{
		return(-1);
	}
	return(16*ms + ls);
}


static int
hexaddr(const char *ptr)
{
	int ms;
	int ls;

	if ((ms = hexbyte(ptr)) < 0)
	{
		return(-1);
	}
	if ((ls = hexbyte(ptr + 2)) < 0)
	{
		return(-1);
	}
	return(256*ms + ls);
}

static int
fileerr(struct ctl *ctl, const char *msg)
{
	fprintf(stderr, "%s:%d:%s\n", ctl->fn, ctl->lineno, msg);
	return(-1);
}


/*
 * Load a single S19 line at ptr into memory
 */
static int
loadline(struct ctl *ctl, const char *ptr)
{
	int len;
	int addr;
	int byte;
	int cksum;
	int tlen;
	unsigned int mask;

	if (*ptr++ != 'S')
	{
		return(fileerr(ctl, "illegal S19 file"));
	}
	if (*ptr++ != '1')
	{
		/* Only interested in S1 records */
		return(0);
	}
	if ((len = hexbyte(ptr)) < 0)
	{
		return(fileerr(ctl, "illegal character"));
	}
	cksum = len;
	ptr += 2;
	tlen = strlen(ptr);
	while (isspace(ptr[tlen - 1]))
	{
		tlen--;
	}
	if (tlen != 2*len)
	{
		return(fileerr(ctl, "line length mismatch"));
	}
	if ((addr = hexaddr(ptr)) < 0)
	{
		return(fileerr(ctl, "illegal character"));
	}
	if (addr > 0xffff || addr < 0x1000)
	{
		return(fileerr(ctl, "illegal address"));
	}
	len -= 2;
	ptr += 4;
	cksum += (addr >> 8) + (addr & 0xff);
	while (len > 1)
	{
		if ((byte = hexbyte(ptr)) < 0)
		{
			return(fileerr(ctl, "illegal character"));
		}
		/*
		 * Taint the affected segment.
		 */
		if ((mask = segment_mask(ctl, addr)) == (unsigned int)(-1))
		{
			return(fileerr(ctl, "illegal byte address"));
		}
		ctl->mask |= mask;
		ctl->mem[addr++] = byte;

		cksum += byte;
		len--;
		ptr += 2;
	}
	cksum = ~cksum & 0xff;
	/*
	 * Now compare the checksums
	 */
	if ((byte = hexbyte(ptr)) < 0)
	{
		return(fileerr(ctl, "illegal character"));
	}
	if (byte != cksum)
	{
		return(fileerr(ctl, "checksum mismatch"));
	}
	return(0);
}

/*
 * Load an S19 config file into memory
 */
static int
loadcfg(struct ctl *ctl)
{
	char buf[256];
	int len;
	int blk;

	ctl->lineno = 1;
	for (;;)
	{
		if ((fgets(buf, sizeof(buf), ctl->cf) != buf))
		{
			if (feof(ctl->cf))
			{

				/*
				 * Which is the first block we expect to be
				 * asked for.
				 */
				for (blk = 0; blk < 255; blk++)
				{
					if (ctl->blk[blk])
					{
						ctl->next = blk;
						break;
					}
				}
				return(0);
			}
			perror(ctl->fn);
			return(-1);
		}
		len = strlen(buf) - 1;
		if (buf[len] != '\n')
		{
			fprintf(stderr, "%s:line %d too long\n", ctl->fn,
					ctl->lineno);
			return(-1);
		}
		buf[len] = '\0';
		if (loadline(ctl, buf) < 0)
		{
			return(-1);
		}
		ctl->lineno++;
	}
}


/*
 * Signal handler.
 */
static void
gotsig(int sig)
{
	terminate = 1;
}

static void
sig_handle(void)
{
	struct sigaction sa;

	sa.sa_flags = 0;
	sigemptyset(&sa.sa_mask);
	sa.sa_handler = gotsig;
	sigaction(SIGINT, &sa, NULL);
}



static int
usage(const char *prog)
{
	fprintf(stderr, "Usage:%s [-p port] <file> [<file>...]\n",
			prog);
	return(-1);
}

int
main(int argc, char **argv)
{
	struct ctl ctl;
	int i;

	ctl.port = "/dev/ttyS0";
	while ((i = getopt(argc, argv, "p:")) != -1)
	{
		switch(i)
		{
		case 'p':
			ctl.port = optarg;
			break;
		default:
			return(usage(argv[0]));
		}
	}
	if (optind == argc)
	{
		return(usage(argv[0]));
	}
	memset(ctl.mem, 0xff, sizeof(ctl.mem));
	memset(ctl.blk, 0, sizeof(ctl.blk));
	ctl.mask = 0;
	ctl.this = 0xff;	/* no segment asked for yet */
	ctl.next = 0xff;	/* no segment availble next */
	while (optind < argc)
	{
		ctl.fn = argv[optind++];
		/*
		 * Load each file nominated into simulated memory.
		 */
		if (!(ctl.cf = fopen(ctl.fn, "r")))
		{
			perror(ctl.fn);
			return(-1);
		}
		/*
		 * Load the config file into memory.
		 */
		if (loadcfg(&ctl) < 0)
		{
			return(-1);
		}
	}
	/*
	 * Open the serial port
	 */
	if ((ctl.serfd = open_serial(ctl.port)) < 0)
	{
		perror(ctl.port);
		return(-1);
	}
	fclose(ctl.cf);
	sig_handle();
	dispatch(&ctl);
	return(0);
}
